/*
    Copyright (C) 2007    Dmitry Ivanov, vonami@gmail.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cstdlib>
#include <iostream>

#include "ext2partition.h"
#include "bitshash.h"
#include "partimage-ng.h"

using namespace std;
using namespace PartimageNG;

Ext2Partition::Ext2Partition(PedPartition *pedPartition)
        : Partition(pedPartition), m_ext2part(0)
{
        errcode_t res;

        res = ext2fs_get_mem(sizeof(struct struct_ext2_filsys), &m_ext2part);
        if ( res ) {
                cerr << "Couldn't allocate memory, res code = " << res << endl;
        }

        res = ext2fs_open(path(), 0, 0, 0, unix_io_manager, &m_ext2part);
        if ( res ) {
                cerr << "Couldn't open the filesystem, res code = " << res << endl;
        }

        res = ext2fs_read_block_bitmap(m_ext2part);
        if ( res ) {
                cerr << "Couldn't read the block bitmap, res code = " << res << endl;
        }

        ext2fs_block_bitmap ext2_bitmap = m_ext2part->block_map;
        m_usedBlocks = m_ext2part->super->s_blocks_count - m_ext2part->super->s_free_blocks_count;

        // if the ext2 bitmap doesn't start with the block #0
        // allocate a new bitmap and set bits for those skipped blocks
        // otherwise we lose data
        // data loss occurs when the data block is 1024 bytes long
        // in this case the ext2 bitmap doesn't cover the very first block
        // (super block) and the bitmap looks like it shifted by 1 bit
        if ( ext2_bitmap->start != 0 ) {
                int full = ext2_bitmap->start >> 3;
                int offset = 0;
                if ( (ext2_bitmap->start & 0x07) != 0 ) {
                        full++;
                        offset = 8 - (ext2_bitmap->start & 0x07);
                }

                m_bitmap.count = m_ext2part->super->s_blocks_count;
                m_bitmap.allocsize = ((ext2_bitmap->real_end + 1) >> 3) + full;
                m_bitmap.bitmap = (u_int8_t *) malloc(m_bitmap.allocsize);
                m_bitmap.offset = offset;

                memcpy(m_bitmap.bitmap + full, ext2_bitmap->bitmap, m_bitmap.allocsize - full);
                memset(m_bitmap.bitmap, 0xFF, full);
                if ( offset != 0 ) {
                        m_bitmap.bitmap[0] = (u_int8_t)(0xFF << offset);
                }
        }
        else {
                m_bitmap.count = m_ext2part->super->s_blocks_count;
                m_bitmap.allocsize = (ext2_bitmap->real_end + 1) >> 3;
                m_bitmap.bitmap = (u_int8_t *) ext2_bitmap->bitmap;
                m_bitmap.offset = 0;
        }

        cerr << "Ext2Partition info:" << endl;
        cerr << "Block size = " << m_ext2part->blocksize << endl;
        cerr << "Blocks count = " << m_ext2part->super->s_blocks_count << endl;
        cerr << "Used blocks count = " << m_usedBlocks << endl;
        cerr << "Bitmap start block = " << ext2_bitmap->start << endl;
        cerr << "Bitmap end block = " << ext2_bitmap->end << endl;
        cerr << "Bitmap allocated size = " << m_bitmap.allocsize << endl;
        cerr << "start = " << ext2_bitmap->start << endl;
        cerr << "end = " << ext2_bitmap->end << endl;
        cerr << "real end = " << ext2_bitmap->real_end << endl;
}

Ext2Partition::~Ext2Partition()
{
        // check if we have allocated memory for the bitmap
        if ( m_ext2part->block_map->start != 0 ) {
                free(m_bitmap.bitmap);
        }

        errcode_t res = ext2fs_close(m_ext2part);
        if ( res ) {
                cerr << "Couldn't close the filesystem, res code = " << res << endl;
        }
}

size_t Ext2Partition::blockSize() const
{
        return m_ext2part->blocksize;
}

u_int64_t Ext2Partition::blocksCount() const
{
        return m_ext2part->super->s_blocks_count;
}

u_int64_t Ext2Partition::usedBlocks() const
{
        return m_usedBlocks;
}

const struct Bitmap* Ext2Partition::bitmap()
{
        return &m_bitmap;
}

bool Ext2Partition::dumpInternalBitmap(int dumpd) const
{
        ext2fs_block_bitmap ext2_bitmap = m_ext2part->block_map;

        u_int32_t bitmapSize = ext2_bitmap->real_end >> 3;
        if ( (ext2_bitmap->real_end & 0x07) != 0 ) {
                bitmapSize++;
        }

        if ( writen(dumpd, ext2_bitmap->bitmap, bitmapSize) == -1 ) {
                perror("Writing the bitmap failed");
                return false;
        }

        return true;
}
