/*
    Copyright (C) 2007    Dmitry Ivanov, vonami@gmail.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cstdlib>
#include <iostream>
#include <fcntl.h>

#include "bitshash.h"
#include "fat32partition.h"
#include "partimage-ng.h"

using namespace std;
using namespace PartimageNG;

// describers the FAT Boot sector, stolen from libparted
struct __attribute__ ((packed)) FatBootSector {
        uint8_t         boot_jump[3];   /* 00: Boot strap short or near jump */
        uint8_t         system_id[8];   /* 03: system name */
        uint16_t        sector_size;    /* 0b: bytes per logical sector */
        uint8_t         cluster_size;   /* 0d: sectors/cluster */
        uint16_t        reserved;       /* 0e: reserved sectors */
        uint8_t         fats;           /* 10: number of FATs */
        uint16_t        dir_entries;    /* 11: number of root directory entries */
        uint16_t        sectors;        /* 13: if 0, total_sect supersedes */
        uint8_t         media;          /* 15: media code */
        uint16_t        fat_length;     /* 16: sectors/FAT for FAT12/16 */
        uint16_t        secs_track;     /* 18: sectors per track */
        uint16_t        heads;          /* 1a: number of heads */
        uint32_t        hidden;         /* 1c: hidden sectors (partition start) */
        uint32_t        sector_count;   /* 20: no. of sectors (if sectors == 0) */

        union __attribute__ ((packed)) {
                /* FAT16 fields */
                struct __attribute__ ((packed)) {
                        uint8_t         drive_num;      /* 24: */
                        uint8_t         empty_1;        /* 25: */
                        uint8_t         ext_signature;  /* 26: always 0x29 */
                        uint32_t        serial_number;  /* 27: */
                        uint8_t         volume_name [11];       /* 2b: */
                        uint8_t         fat_name [8];   /* 36: */
                        uint8_t         boot_code[448]; /* 3f: Boot code (or message) */
                } fat16;
                /* FAT32 fields */
                struct __attribute__ ((packed)) {
                        uint32_t        fat_length;     /* 24: size of FAT in sectors */
                        uint16_t        flags;          /* 28: bit8: fat mirroring, low4: active fat */
                        uint16_t        version;        /* 2a: minor * 256 + major */
                        uint32_t        root_dir_cluster;       /* 2c: */
                        uint16_t        info_sector;    /* 30: */
                        uint16_t        backup_sector;  /* 32: */
                        uint8_t         empty_1 [12];   /* 34: */
                        uint16_t        drive_num;      /* 40: */
                        uint8_t         ext_signature;  /* 42: always 0x29 */
                        uint32_t        serial_number;  /* 43: */
                        uint8_t         volume_name [11];       /* 47: */
                        uint8_t         fat_name [8];   /* 52: */
                        uint8_t         boot_code[420]; /* 5a: Boot code (or message) */
                } fat32;
        } u;

        uint16_t        boot_sign;      /* 1fe: always 0xAA55 */
};

Fat32Partition::Fat32Partition(PedPartition *pedPartition)
        : Partition(pedPartition)
{
        // read the fat boot sector
        struct FatBootSector bs;
        int devd = open(path(), O_RDONLY);
        if ( devd == -1 ) {
                perror("Couldn't open a device file");
                return;
        }
        if ( readn(devd, &bs, sizeof(FatBootSector)) < sizeof(FatBootSector) ) {
                cerr << "Reading the FAT boot sector failed" << endl;
                return;
        }

        // calculate and print some important values
        u_int32_t clusterOffset = bs.reserved + bs.fats *
                                   bs.u.fat32.fat_length;
        u_int32_t clusterCount = (bs.sector_count - clusterOffset) / bs.cluster_size;
        u_int32_t clusterSize = bs.cluster_size;
        cerr << "Fat32Partition info:" << endl;
        cerr << "Total sectors: " << bs.sector_count << endl;
        cerr << "Sectors allocation: [" << bs.reserved
                                  << " + " << (int)bs.fats
                                  << "fats * " <<  bs.u.fat32.fat_length
                                  << " + " << clusterCount
                                  << "cls * " << clusterSize
                                  << "] = " << clusterOffset  + clusterCount * clusterSize
                                  << endl;

        // calculate a bitmap for the reserved + FAT sectors
        int full = clusterOffset >> 3;
        int offset = 0;
        if ( (clusterOffset & 0x07) != 0 ) {
                full++;
                offset = 8 - (clusterOffset & 0x07);
        }

        // generate the bitmap and print some stats
        m_bitmap.offset = offset;
        m_bitmap.count = clusterOffset + clusterCount * clusterSize;
        m_bitmap.allocsize = full + ((clusterCount * clusterSize) >> 3);
        if ( ((clusterCount * clusterSize) & 0x07) != 0 ) {
                m_bitmap.allocsize++;
        }
        m_bitmap.bitmap = (u_int8_t *) malloc(m_bitmap.allocsize);
        memset(m_bitmap.bitmap, 0x00, m_bitmap.allocsize);
        memset(m_bitmap.bitmap, 0xFF, full);
        if ( offset != 0 ) {
                m_bitmap.bitmap[0] = (u_int8_t)(0xFF << offset);
        }

        cerr << "Bitmap start block: " << m_bitmap.offset << endl;
        cerr << "Bitmap block count: " << m_bitmap.count << endl;
        cerr << "Bitmap allocated size: " << m_bitmap.allocsize << endl;

        // read off the FAT and convert it
        m_fatSize = bs.u.fat32.fat_length * 512;
        m_fat = (u_int32_t *) malloc(m_fatSize);
        if ( preadn(devd, m_fat, m_fatSize, bs.reserved * 512) < m_fatSize) {
                cerr << "Reading the FAT table failed" << endl;
                return;
        }

        u_int32_t currentCluster = 0;
        u_int32_t currentBitmapByte = full;     // skip the reserved sectors + FAT sectors
        u_int8_t byte = 0;
        if ( clusterSize < 8 ) {
                while ( currentCluster < clusterCount ) {
                        u_int8_t i = 0;
                        u_int8_t mask = 0xFF >> (8 - clusterSize);
                        while ( (i < 8/clusterSize) && (currentCluster < clusterCount) ) {
                                if ( !fatClusterFree(m_fat, currentCluster) ) {
                                        byte = byte | mask;
                                }
                                mask = mask << clusterSize;
                                i++;
                                currentCluster++;
                        }
                        m_bitmap.bitmap[currentBitmapByte] = byte;
                        currentBitmapByte++;
                }
        }
        else {
                while ( currentCluster < clusterCount ) {
                        if ( !fatClusterFree(m_fat, currentCluster) )
                                byte = 0xFF;
                        else
                                byte = 0x00;

                        for (u_int8_t i = 0; i < clusterSize/8; ++i ) {
                                m_bitmap.bitmap[currentBitmapByte] = byte;
                                currentBitmapByte++;
                        }
                        currentCluster++;
                }
        }

        //  walk through the bitmap and count up every used block
        m_usedBlocks = 0;
        for (u_int64_t i = 0; i < m_bitmap.allocsize; ++i ) {
                m_usedBlocks += bits_hash[(u_int8_t)m_bitmap.bitmap[i]];
        }

        cerr << "Block size = " << 512 << endl;
        cerr << "Blocks count = " << m_bitmap.count << endl;
        cerr << "Used blocks count = " << m_usedBlocks << endl;
}

Fat32Partition::~Fat32Partition()
{
        free(m_fat);
        free(m_bitmap.bitmap);
}

size_t Fat32Partition::blockSize() const
{
        return 512;
}

u_int64_t Fat32Partition::blocksCount() const
{
        return m_bitmap.count;
}

u_int64_t Fat32Partition::usedBlocks() const
{
        return m_usedBlocks;
}

const struct Bitmap* Fat32Partition::bitmap()
{
        return &m_bitmap;
}

bool Fat32Partition::fatClusterFree(u_int32_t *fat, u_int32_t ncluster) const
{
        switch ( fat[2 + ncluster] & 0x0fffffff ) {
                case 0x00000000:
                        // free
                        return true;
                        break;
                case 0x0ffffff7:
                        // bad/reserved
                        return true;
                        break;
                default:
                        return false;
        }
}

bool Fat32Partition::dumpInternalBitmap(int dumpd) const
{
        if ( writen(dumpd, m_fat, m_fatSize) == -1 ) {
                perror("Writing the bitmap failed");
                return false;
        }

        return true;
}
