/*
    Copyright (C) 2007    Dmitry Ivanov, vonami@gmail.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <fcntl.h>

#include "imager.h"
#include "partitionmanager.h"
#include "endian.h"

extern "C" {
        #include <parted/constraint.h>
        #include <parted/device.h>
        #include <parted/disk.h>
}

using namespace std;
using namespace PartimageNG;

Imager::Imager():
        m_bufferSize(0), m_buffer(0),
        m_saveMbr(true), m_grubHack(true),
        m_saveFullSwap(false), m_descriptionSize(255),
        m_ioErrorsHandling(StopIOErrors)
{
        m_vectorCtx.count = 0;
        m_vectorCtx.size = 0;
        m_vectorCtx.vectors = 0;
        setDesiredBuffering(1);
        m_description = (char *) malloc(m_descriptionSize + 1);
        memset(m_description, 0, m_descriptionSize + 1);
}

Imager::~Imager()
{
        free(m_buffer);
        free(m_vectorCtx.vectors);
}

void Imager::setSaveMbr(bool saveMbr)
{
        m_saveMbr = saveMbr;
}

bool Imager::saveMbr() const
{
        return m_saveMbr;
}

void Imager::setGrubHack(bool grubHack)
{
        m_grubHack = grubHack;
}

bool Imager::grubHack() const
{
        return m_grubHack;
}

void Imager::setSaveFullSwap(bool saveFullSwap)
{
        m_saveFullSwap = saveFullSwap;
}

bool Imager::saveFullSwap() const
{
        return m_saveFullSwap;
}

void Imager::setDesiredBuffering(u_int16_t desiredBuffering)
{
        m_desiredBuffering = desiredBuffering;

        // preallocate m_vectorCtx to hold at least
        // 'desiredBuffering' number of vectors
        if ( m_vectorCtx.size < m_desiredBuffering ) {
                m_vectorCtx.size = m_desiredBuffering;
                m_vectorCtx.vectors = (struct Vector *) realloc(m_vectorCtx.vectors,
                                                                sizeof(struct Vector) * m_desiredBuffering);
        }
}

u_int16_t Imager::desiredBuffering() const
{
        return m_desiredBuffering;
}

void Imager::setDescription(const char *description)
{
        if ( m_descriptionSize < strlen(description) ) {
                m_descriptionSize = strlen(description);
                m_description = (char *) realloc(m_description, m_descriptionSize + 1);
                memset(m_description, 0, m_descriptionSize + 1);
        }

        strncpy(m_description, description, m_descriptionSize);
}

const char* Imager::description() const
{
        return m_description;
}

void Imager::setIOErrorsHandling(IOErrorsHandling ioErrorsHandling)
{
        m_ioErrorsHandling = ioErrorsHandling;
}

IOErrorsHandling Imager::ioErrorsHandling() const
{
        return m_ioErrorsHandling;
}

void Imager::allocateBuffer(size_t blockSize)
{
        // calculate required buffer size
        // TODO: need to put additional logic to determine the most
        // appropriate buffering for the system and avoid buffer overflows
        size_t bufferSize = m_desiredBuffering * blockSize;
        if ( bufferSize > m_bufferSize ) {
                m_bufferSize = bufferSize;
                m_buffer = (u_int8_t *) realloc(m_buffer, m_bufferSize);
        }
}

u_int64_t Imager::vectorize(const struct Bitmap *bitmap, u_int64_t start, u_int16_t count)
{
        if ( m_vectorCtx.size < count ) {
                m_vectorCtx.size = count;
                m_vectorCtx.vectors = (struct Vector *) realloc(m_vectorCtx.vectors, sizeof(struct Vector) * count);
        }

        u_int64_t current = start;
        u_int16_t vectorized = 0;
        u_int64_t handled = 0;
        m_vectorCtx.count = 0;
        Vector *vector = m_vectorCtx.vectors;
        vector->nblock = 0;
        vector->count = 0;
        while ( current < bitmap->count && vectorized < count ) {
                if ( bitmap->bitmap[(current + bitmap->offset) >> 3] & (1 << ((current + bitmap->offset) & 0x07)) ) {
                        if ( vector->count == 0 ) {
                                vector->nblock = current;
                                vector->count = 1;
                        }
                        else {
                                vector->count++;
                        }
                        vectorized++;
                }
                else {
                        if ( vector->count > 0 ) {
                                vector++;
                                m_vectorCtx.count++;
                                vector->nblock = 0;
                                vector->count = 0;
                        }
                        else {
                                //  ignore zeroes
                        }
                }
                current++;
                handled++;
        }

        if ( vector->count > 0 ) {
                m_vectorCtx.count++;
        }

        return handled;
}

void Imager::dumpVectorCtx()
{
        cerr << "VectorCtx.count = " << m_vectorCtx.count;
        cerr << "VectorCtx.size = " << m_vectorCtx.size;
        for ( u_int16_t i = 0; i < m_vectorCtx.count; ++i ) {
                cerr << "vector[" << i << "] = from " << m_vectorCtx.vectors[i].nblock;
                cerr << ", count " << m_vectorCtx.vectors[i].count << endl;
        }
}

int Imager::readVector(int devd, size_t blockSize)
{
        u_int8_t *buf = m_buffer;
        off_t offset = 0;
        size_t chunkSize = 0;
        ssize_t readnum = 0;
        for ( u_int16_t i = 0; i < m_vectorCtx.count; ++i ) {
                offset = m_vectorCtx.vectors[i].nblock * blockSize;
                chunkSize = m_vectorCtx.vectors[i].count * blockSize;

                if ( (readnum = preadn(devd, buf, chunkSize, offset, m_ioErrorsHandling)) == -1 ) {
                        perror("Reading data block failed");
                        return 1;
                }
                else if ( readnum < chunkSize ) {
                        cerr << "Read incomplete data block" << endl;
                }
                buf += chunkSize;
        }

        return 0;
}

int Imager::writeVector(int outd, size_t blockSize)
{
        u_int8_t *buf = m_buffer;
        off_t offset = 0;
        size_t chunkSize = 0;
        ssize_t writenum = 0;
        for ( u_int16_t i = 0; i < m_vectorCtx.count; ++i ) {
                offset = m_vectorCtx.vectors[i].nblock * blockSize;
                chunkSize = m_vectorCtx.vectors[i].count * blockSize;

                if ( (writenum = pwriten(outd, buf, chunkSize, offset, m_ioErrorsHandling)) == -1 ) {
                        perror("Writing data block failed");
                        return 1;
                }
                else if ( writenum < chunkSize ) {
                        cerr << "Written incomplete data block" << endl;
                }
                buf += chunkSize;
        }

        return 0;
}

int Imager::savePartition(const char *ifile, const char *ofile)
{
        // open the output file
        int outd = -1;
        if ( strcmp(ofile, "stdout") == 0 ) {
                outd = STDOUT_FILENO;
        }
        else {
                outd = open(ofile, O_CREAT | O_EXCL | O_WRONLY, 0644);
                if ( outd == -1 ) {
                        perror("Couldn't create an image file");
                        return 1;
                }
        }

        // allocate a Partition
        PartitionManager::InstantiateOption opt = PartitionManager::DetectFilesystem;
        if ( PartitionManager::filesystem(ifile) != 0 &&
                strcmp(PartitionManager::filesystem(ifile), "linux-swap") == 0 &&
                m_saveFullSwap ) {
                opt = PartitionManager::ForceUnknownFilesystem;
        }

        Partition *part = PartitionManager::instantiatePartition(ifile, opt);
        if ( !part ) {
                cerr << "The specified partition \'" << ifile << "\' was not found on the disk" << endl;
                return 1;
        }

        // make up the image header
        ImageHeader imageHeader;
        memset(&imageHeader, 0, sizeof(ImageHeader));
        strncpy(imageHeader.magic, PARTIMAGE_NG_MAGIC, strlen(PARTIMAGE_NG_MAGIC));
        imageHeader.version = cpu_img16(0);
        imageHeader.type = PARTIMAGE_NG_PARTITION_IMAGE;

        // convert description
        char *desc_buf = NULL;
        size_t desc_size = 0;
        if ( convertToUTF8(&desc_buf, &desc_size, m_description, strlen(m_description), nl_langinfo(CODESET)) != 0 ) {
                cerr << "Failed to convert the description to UTF-8" << endl;
                desc_size = 0;
        }
        imageHeader.descriptionSize = cpu_img16(desc_size);

        // write the umage header and description
        if ( writen(outd, &imageHeader, sizeof(ImageHeader)) == -1 ) {
                perror("Writing the image header failed");
                return 1;
        }

        if ( writen(outd, desc_buf, desc_size) == -1 ) {
                perror("Writing description failed");
                return 1;
        }

        // finally, save partition's data
        savePartitionSnapshot(part, outd);

        free(desc_buf);
        delete part;
        if ( outd != STDOUT_FILENO )
                close(outd);

        return 0;
}

int Imager::restorePartition(const char *ifile, const char *ofile)
{
        //  open the image file
        int imaged = -1;
        if ( strcmp(ifile, "stdin") == 0 ) {
                imaged = STDIN_FILENO;
        }
        else {
                imaged = open(ifile, O_RDONLY);
                if ( imaged == -1 ) {
                        perror("Couldn't open the image file");
                        return 1;
                }
        }

        //  read the image header
        ImageHeader imageHeader;
        if ( readn(imaged, &imageHeader, sizeof(ImageHeader)) != sizeof(ImageHeader) ) {
                perror("Reading header failed");
                return 1;
        }

        if ( imageHeader.type != PARTIMAGE_NG_PARTITION_IMAGE ) {
                cerr << "File " << ifile << " is not an image of a partition" << endl;
                return 1;
        }

        // read the description
        u_int16_t desc_size = img16_cpu(imageHeader.descriptionSize);
        char *desc_buf = (char *) malloc(desc_size);
        if ( readn(imaged, desc_buf, desc_size) != desc_size ) {
                perror("Reading the description failed");
                return 1;
        }

        // finally, restore the partition
        restorePartitionSnapshot(imaged, ofile, UseExistingPartition);

        free(desc_buf);
        if ( imaged != STDIN_FILENO )
                close(imaged);

        return 0;
}

int Imager::saveDisk(const char *ifile, const char *ofile)
{
        // open the output file
        int outd = -1;
        if ( strcmp(ofile, "stdout") == 0 ) {
                outd = STDOUT_FILENO;
        }
        else {
                outd = open(ofile, O_CREAT | O_EXCL | O_WRONLY, 0644);
                if ( outd == -1 ) {
                        perror("Couldn't create an image file");
                        return 1;
                }
        }

        // get the list of partitions
        list<string> parts = PartitionManager::partitionsList(ifile);
        cerr << " == Found " << parts.size() << " partitions on the disk, " <<
                "partition table type: \'" << PartitionManager::labelname(ifile) << "\' == " << endl;

        // make up the image header
        ImageHeader imageHeader;
        memset(&imageHeader, 0, sizeof(ImageHeader));
        strncpy(imageHeader.magic, PARTIMAGE_NG_MAGIC, strlen(PARTIMAGE_NG_MAGIC));
        imageHeader.version = cpu_img16(0);
        imageHeader.type = PARTIMAGE_NG_DISK_IMAGE;
        imageHeader.partitionsCount = cpu_img16(parts.size());
        imageHeader.labelnameSize = cpu_img16(strlen(PartitionManager::labelname(ifile)));

        // convert description
        char *desc_buf = NULL;
        size_t desc_size = 0;
        if ( convertToUTF8(&desc_buf, &desc_size, m_description, strlen(m_description), nl_langinfo(CODESET)) != 0 ) {
                cerr << "Failed to convert the description to UTF-8" << endl;
                desc_size = 0;
        }
        imageHeader.descriptionSize = cpu_img16(desc_size);

        // count up the raw data blocks we are going to save
        if ( m_saveMbr )
                imageHeader.rawdataCount = cpu_img16( img16_cpu(imageHeader.rawdataCount) + 1 );

        if ( m_grubHack )
                imageHeader.rawdataCount = cpu_img16( img16_cpu(imageHeader.rawdataCount) + 1 );

        if ( writen(outd, &imageHeader, sizeof(ImageHeader)) == -1 ) {
                perror("Writing the image header failed");
                return 1;
        }
        if ( writen(outd, desc_buf, desc_size) == -1 ) {
                perror("Writing the description failed");
                return 1;
        }
        if ( writen(outd, PartitionManager::labelname(ifile), img16_cpu(imageHeader.labelnameSize)) == -1 ) {
                perror("Writing the partition table type failed");
                return 1;
        }

        // save mbr and the grub boot code if requested
        if ( m_saveMbr || m_grubHack ) {
                int devd = open(ifile, O_RDONLY);
                if ( devd == -1 ) {
                        perror("Couldn't open a device file");
                        return 1;
                }

                RawdataHeader rawdataHeader;
                if ( m_saveMbr ) {
                        u_int8_t mbr[446];
                        if ( preadn(devd, &mbr, sizeof(mbr), 0) < sizeof(mbr) ) {
                                perror("Reading MBR failed");
                                return 1;
                        }
                        else {
                                rawdataHeader.offset = cpu_img64(0);
                                rawdataHeader.size = cpu_img64(sizeof(mbr));
                                if ( writen(outd, &rawdataHeader, sizeof(RawdataHeader)) == -1 ) {
                                        perror("Writing the rawdata header failed");
                                        return 1;
                                }
                                if ( writen(outd, &mbr, sizeof(mbr)) == -1 ) {
                                        perror("Writing the mbr failed");
                                        return 1;
                                }
                        }
                }

                if ( m_grubHack ) {
                        // grub stores stage 1.5 in the 62 spare sectors
                        // right after the MBR
                        u_int8_t grub[512 * 62];
                        if ( preadn(devd, &grub, sizeof(grub), 512) < sizeof(grub) ) {
                                perror("Reading GRUB boot code failed");
                                return 1;
                        }
                        else {
                                rawdataHeader.offset = cpu_img64(512);
                                rawdataHeader.size = cpu_img64(sizeof(grub));
                                if ( writen(outd, &rawdataHeader, sizeof(RawdataHeader)) == -1 ) {
                                        perror("Writing the rawdata header failed");
                                        return 1;
                                }
                                if ( writen(outd, &grub, sizeof(grub)) == -1 ) {
                                        perror("Writing the GRUB boot code failed");
                                        return 1;
                                }
                        }
                }

                close(devd);
        }

        // walk over the partitions and save them
        for ( list<string>::iterator it = parts.begin(); it != parts.end(); it++ ) {
                PartitionManager::InstantiateOption opt = PartitionManager::DetectFilesystem;
                if ( PartitionManager::filesystem((*it).c_str()) != 0 &&
                     strcmp(PartitionManager::filesystem((*it).c_str()), "linux-swap") == 0 &&
                     m_saveFullSwap ) {
                        opt = PartitionManager::ForceUnknownFilesystem;
                }
                Partition *part = PartitionManager::instantiatePartition((*it).c_str(), opt);
                savePartitionSnapshot(part, outd);
                delete part;
        }

        free(desc_buf);
        if ( outd != STDOUT_FILENO )
                close(outd);

        return 0;
}

int Imager::savePartitionSnapshot(Partition *part, int outd)
{
        //  open the partition for reading data
        int devd = open(part->path(), O_RDONLY);
        if ( devd == -1 ) {
                perror("Couldn't open the device file");
                return 1;
        }

        // make up the partition header
        PartitionHeader partitionHeader;
        memset(&partitionHeader, 0, sizeof(PartitionHeader));
        partitionHeader.type = cpu_img16(part->type());
        partitionHeader.startSector = cpu_img64(part->startSector());
        partitionHeader.endSector = cpu_img64(part->endSector());

        // for extended partitions the header contains only partition geometry
        // (besides the partition type, obviously)
        if ( img16_cpu(partitionHeader.type) == PED_PARTITION_EXTENDED ) {
                if ( writen(outd, &partitionHeader, sizeof(PartitionHeader)) == -1 ) {
                        perror("Writing the partition header failed");
                        return 1;
                }

                close(devd);
                return 0;
        }

        const Bitmap *bitmap = part->bitmap();
        partitionHeader.blockSize = cpu_img64(part->blockSize());
        partitionHeader.blocksCount = cpu_img64(part->blocksCount());
        partitionHeader.usedBlocks = cpu_img64(part->usedBlocks());
        if ( part->filesystem() != 0 )
                partitionHeader.fsnameSize = cpu_img16(strlen(part->filesystem()));
        else
                partitionHeader.fsnameSize = cpu_img16(0);
        partitionHeader.bitmapSize = cpu_img64(bitmap->allocsize);
        partitionHeader.bitmapOffset = cpu_img64(bitmap->offset);

        // write the partition header
        if ( writen(outd, &partitionHeader, sizeof(PartitionHeader)) == -1 ) {
                perror("Writing the partition header failed");
                return 1;
        }

        // write the filesystem name
        if ( writen(outd, part->filesystem(), img16_cpu(partitionHeader.fsnameSize)) == -1 ) {
                perror("Writing the filesystem name failed");
                return 1;
        }

        // write the bitmap
        if ( writen(outd, bitmap->bitmap, img64_cpu(partitionHeader.bitmapSize)) == -1 ) {
                perror("Writing the bitmap failed");
                return 1;
        }

        //  allocate internal buffer
        allocateBuffer(img64_cpu(partitionHeader.blockSize));

        //  walk over the bitmap and read blocks
        u_int64_t blocksCount = img64_cpu(partitionHeader.blocksCount);
        u_int64_t usedBlocks = img64_cpu(partitionHeader.usedBlocks);
        u_int64_t blockSize = img64_cpu(partitionHeader.blockSize);
        u_int64_t currentBlock = 0;
        u_int16_t actualBuffering = 0;
        u_int64_t total = 0;
        cerr << "Saving block #: ";
        while ( currentBlock < blocksCount ) {
                cerr << "\033[s" << total << " (" << ((float)(total) / usedBlocks) * 100 << "%)";

                currentBlock += vectorize(bitmap, currentBlock, m_desiredBuffering);

                actualBuffering = 0;
                for ( int i = 0; i < m_vectorCtx.count; ++i ) {
                        actualBuffering += m_vectorCtx.vectors[i].count;
                }
                if ( actualBuffering > m_desiredBuffering ) {
                        cerr << "Imager::createImage(): internal buffer overflow" << endl;
                        dumpVectorCtx();
                        return 1;
                }
                total += actualBuffering;

                if ( readVector(devd, blockSize) == 1 ) {
                        perror("Reading from device failed");
                        return 1;
                }
                if ( writen(outd, m_buffer, actualBuffering * blockSize) == -1 ) {
                        perror("Writing to the image file failed");
                        return 1;
                }

                cerr << "\033[K" << "\033[u";
        }
        cerr << "Number of blocks saved: " << total << endl;

        close(devd);
        return 0;
}

int Imager::restoreDisk(const char *ifile, const char *ofile)
{
        //  open the image file
        int imaged = -1;
        if ( strcmp(ifile, "stdin") == 0 ) {
                imaged = STDIN_FILENO;
        }
        else {
                imaged = open(ifile, O_RDONLY);
                if ( imaged == -1 ) {
                        perror("Couldn't open the image file");
                        return 1;
                }
        }

        //  read the header
        ImageHeader imageHeader;
        if ( readn(imaged, &imageHeader, sizeof(ImageHeader)) != sizeof(ImageHeader) ) {
                perror("Reading header failed");
                return 1;
        }

        if ( imageHeader.type != PARTIMAGE_NG_DISK_IMAGE ) {
                cerr << "File " << ifile << " is not an image of a drive" << endl;
                return 1;
        }

        // read the description
        u_int16_t desc_size = img16_cpu(imageHeader.descriptionSize);
        char *desc_buf = (char *) malloc(desc_size);
        if ( readn(imaged, desc_buf, desc_size) != desc_size ) {
                perror("Reading the description failed");
                return 1;
        }

        // read the disk label
        u_int16_t labelnameSize = img16_cpu(imageHeader.labelnameSize);
        char *labelname = (char *) malloc(labelnameSize + 1);
        labelname[labelnameSize] = '\0';
        if ( readn(imaged, labelname, labelnameSize) != labelnameSize ) {
                perror("Reading disk label failed");
                return 1;
        }

        cerr << "Partition table type: " << labelname << endl;
        if ( PartitionManager::createLabel(ofile, labelname) != 0) {
                cerr << "Couldn't create a new partition table on the drive" << endl;
                return 1;
        }

        // restore raw data blocks if any
        size_t rawdataSize = 0;
        u_int8_t *rawdata = 0;
        int devd = -1;
        for ( int i = 0; i < img16_cpu(imageHeader.rawdataCount); ++i ) {
                RawdataHeader rawdataHeader;
                if ( readn(imaged, &rawdataHeader, sizeof(rawdataHeader)) == -1 ) {
                        perror("Reading the rawdata header failed");
                        return 1;
                }

                if ( rawdataSize < img64_cpu(rawdataHeader.size) ) {
                        rawdataSize = img64_cpu(rawdataHeader.size);
                        rawdata = (u_int8_t *) realloc(rawdata, rawdataSize);
                }

                if ( readn(imaged, rawdata, rawdataSize) == -1 ) {
                        perror("Reading the rawdata failed");
                        return 1;
                }

                if ( devd == -1 ) {
                        devd = open(ofile, O_WRONLY);
                        if ( devd == -1 ) {
                                perror("Couldn't open a device file");
                                return 1;
                        }
                }

                if ( pwriten(devd, rawdata, rawdataSize, img64_cpu(rawdataHeader.offset)) == -1 ) {
                        perror("Writing a rawdata block failed");
                        return 1;
                }
        }
        free(rawdata);
        if ( devd != -1 )
                close(devd);


        // walk over the saved partitions and restore them
        for ( int i = 0; i < img16_cpu(imageHeader.partitionsCount); ++i ) {
                if ( restorePartitionSnapshot(imaged, ofile, CreateNewPartition) != 0 ) {
                        // clean up
                        // TODO: using goto: a better code practice?
                        free(labelname);
                        if ( imaged != STDIN_FILENO )
                                close(imaged);
                        return 1;
                }
        }

        free(desc_buf);
        free(labelname);
        if ( imaged != STDIN_FILENO )
                close(imaged);

        return 0;
}

int Imager::restorePartitionSnapshot(int imaged, const char *ofile, RestorePartitionMode mode)
{
        // read the partition header
        PartitionHeader header;
        if ( readn(imaged, &header, sizeof(PartitionHeader)) != sizeof(PartitionHeader) ) {
                perror("Reading partition header failed");
                return 1;
        }

        // convert the header: image byte order -> cpu byte order
        // will be much easier to use
        header.type = img16_cpu(header.type);
        header.startSector = img64_cpu(header.startSector);
        header.endSector = img64_cpu(header.endSector);
        header.blockSize = img64_cpu(header.blockSize);
        header.blocksCount = img64_cpu(header.blocksCount);
        header.usedBlocks = img64_cpu(header.usedBlocks);
        header.fsnameSize = img16_cpu(header.fsnameSize);
        header.bitmapSize = img64_cpu(header.bitmapSize);
        header.bitmapOffset = img64_cpu(header.bitmapOffset);

        cerr << " == Processing partition: type " << header.type << ", geometry {"
                << header.startSector << ", " << header.endSector << "} == " << endl;

        // read the filesystem name
        char *fsname = 0;
        if ( header.type != PED_PARTITION_EXTENDED && header.fsnameSize != 0 ) {
                fsname = (char *) malloc(header.fsnameSize + 1);
                fsname[header.fsnameSize] = '\0';
                if ( readn(imaged, fsname, header.fsnameSize) != header.fsnameSize ) {
                        perror("Reading filesystem name failed");
                        return 1;
                }

                cerr << "Filesystem: " << fsname << endl;
        }

        cerr << "Block size = " << header.blockSize << endl;
        cerr << "Number of blocks to restore = " << header.blocksCount << endl;
        cerr << "Number of used blocks to restore = " << header.usedBlocks << endl;
        cerr << "Bitmap allocated size = " << header.bitmapSize << endl;

        int devd = -1;
        if (  mode == CreateNewPartition ) {
                const char *opath = PartitionManager::createPartition(ofile, header.type, fsname,
                                                                      header.startSector, header.endSector);

                if ( opath == 0 ) {
                        cerr << "Failed to create a new partition on the disk" << endl;
                        cerr << "Check if you have enough permissions to do this" << endl;
                        return 1;
                }

                if ( header.type == PED_PARTITION_EXTENDED ) {
                        cerr << "Restored an extended partition" << endl;
                        free(fsname);
                        return 0;
                }
                else {
                        devd = open(opath, O_WRONLY);
                        if ( devd == -1 ) {
                                perror("Couldn't open a device file");
                                return 1;
                        }
                }
        }
        else {
                devd = open(ofile, O_WRONLY);
                if ( devd == -1 ) {
                        perror("Couldn't open a device file");
                        return 1;
                }
        }

        //  read the bitmap
        struct Bitmap bitmap;
        bitmap.count = header.blocksCount;
        bitmap.allocsize = header.bitmapSize;
        bitmap.bitmap = (u_int8_t *) malloc(bitmap.allocsize);
        bitmap.offset = header.bitmapOffset;
        if ( readn(imaged, bitmap.bitmap, bitmap.allocsize) != bitmap.allocsize ) {
                perror("Reading bitmap failed");
                return 1;
        }

        //  allocate internal buffer
        allocateBuffer(header.blockSize);

        //  walk over the bitmap and write blocks
        u_int64_t currentBlock = 0;
        u_int16_t actualBuffering = 0;
        u_int64_t total = 0;
        cerr << "Restoring block #: ";
        while ( currentBlock < header.blocksCount ) {
                cerr << "\033[s" << total << " (" << ((float)(total) / header.usedBlocks) * 100 << "%)";

                currentBlock += vectorize(&bitmap, currentBlock, m_desiredBuffering);

                actualBuffering = 0;
                for ( int i = 0; i < m_vectorCtx.count; ++i ) {
                        actualBuffering += m_vectorCtx.vectors[i].count;
                }
                if ( actualBuffering > m_desiredBuffering ) {
                        cerr << "Imager::restoreImage(): internal buffer overflow" << endl;
                        dumpVectorCtx();
                        return 1;
                }
                total += actualBuffering;

                if ( readn(imaged, m_buffer, actualBuffering * header.blockSize) != actualBuffering * header.blockSize ) {
                        perror("Reading the image file failed");
                        return 1;
                }
                if ( writeVector(devd, header.blockSize) == 1 ) {
                        perror("Writing to device failed");
                        return 1;
                }

                cerr << "\033[K" << "\033[u";
        }
        cerr << "Number of blocks restored: " << total << endl;

        free(fsname);
        free(bitmap.bitmap);
        close(devd);
        return 0;
}

int Imager::imageinfo(const char *imgfile) const
{
        //  open the image file
        int imaged = -1;
        imaged = open(imgfile, O_RDONLY);
        if ( imaged == -1 ) {
                perror("Couldn't open the image file");
                return 1;
        }

        //  read the header
        ImageHeader imageHeader;
        if ( readn(imaged, &imageHeader, sizeof(ImageHeader)) != sizeof(ImageHeader) ) {
                perror("Reading header failed");
                return 1;
        }

        imageHeader.version = img16_cpu(imageHeader.version);
        imageHeader.partitionsCount = img16_cpu(imageHeader.partitionsCount);
        imageHeader.descriptionSize = img16_cpu(imageHeader.descriptionSize);
        imageHeader.labelnameSize = img16_cpu(imageHeader.labelnameSize);
        imageHeader.rawdataCount = img16_cpu(imageHeader.rawdataCount);

        if ( strcmp(imageHeader.magic, PARTIMAGE_NG_MAGIC) != 0 ) {
                cerr << "This file is not a correct partimage-ng image file" << endl;
                return 1;
        }

        cerr << "Version: " << imageHeader.version << endl;

        if ( imageHeader.type == PARTIMAGE_NG_DISK_IMAGE ) {
                cerr << "Type: " << "image of a disk" << endl;
        }
        else if (imageHeader.type == PARTIMAGE_NG_PARTITION_IMAGE ) {
                cerr << "Type: " << "image of a partition" << endl;
        }
        else {
                cerr << "Type: " << "unknown image type " << imageHeader.type << endl;
                return 1;
        }

        // read the description
        u_int16_t desc_size = imageHeader.descriptionSize;
        char *desc_buf = (char *) malloc(desc_size);
        if ( readn(imaged, desc_buf, desc_size) != desc_size ) {
                perror("Reading the description failed");
                return 1;
        }
        char *conv_buf = NULL;
        size_t conv_size = 0;
        if ( convertFromUTF8(&conv_buf, &conv_size, desc_buf, desc_size, nl_langinfo(CODESET)) != 0 ) {
                cerr << "Failed to convert the description from UTF-8" << endl;
        }
        else {
                cerr << "Description: " << conv_buf << endl;
        }
        free(desc_buf);
        free(conv_buf);

        if ( imageHeader.type == PARTIMAGE_NG_DISK_IMAGE ) {
                u_int16_t labelnameSize = imageHeader.labelnameSize;
                char *labelname = (char *) malloc(labelnameSize + 1);
                labelname[labelnameSize] = '\0';
                if ( readn(imaged, labelname, labelnameSize) != labelnameSize ) {
                        perror("Reading disk label failed");
                        return 1;
                }

                cerr << "Partition table: " << labelname << endl;
                cerr << "Partitions: " << imageHeader.partitionsCount << endl;
                free(labelname);

                // read raw data blocks if any
                for ( int i = 0; i < imageHeader.rawdataCount; ++i ) {
                        RawdataHeader rawdataHeader;
                        if ( readn(imaged, &rawdataHeader, sizeof(rawdataHeader)) == -1 ) {
                                perror("Reading the rawdata header failed");
                                return 1;
                        }

                        cerr << "Rawdata: " << img64_cpu(rawdataHeader.size) << " bytes at "
                                            << img64_cpu(rawdataHeader.offset) << endl;
                        if ( lseek(imaged, img64_cpu(rawdataHeader.size), SEEK_CUR) == -1 ) {
                                perror("Skipping the rawdata failed");
                                return 1;
                        }
                }

                for ( int i = 0; i < imageHeader.partitionsCount; i++ ) {
                        // read the partition header
                        PartitionHeader header;
                        if ( readn(imaged, &header, sizeof(PartitionHeader)) != sizeof(PartitionHeader) ) {
                                perror("Reading partition header failed");
                                return 1;
                        }

                        // convert the header: image byte order -> cpu byte order
                        // will be much easier to use
                        header.type = img16_cpu(header.type);
                        header.startSector = img64_cpu(header.startSector);
                        header.endSector = img64_cpu(header.endSector);
                        header.blockSize = img64_cpu(header.blockSize);
                        header.blocksCount = img64_cpu(header.blocksCount);
                        header.usedBlocks = img64_cpu(header.usedBlocks);
                        header.fsnameSize = img16_cpu(header.fsnameSize);
                        header.bitmapSize = img64_cpu(header.bitmapSize);
                        header.bitmapOffset = img64_cpu(header.bitmapOffset);

                        cerr << " == Partition: type " << header.type << ", geometry {"
                                << header.startSector << ", " << header.endSector << "} == " << endl;

                        if ( header.type == PED_PARTITION_EXTENDED )
                                continue;

                        // read the filesystem name
                        char *fsname = 0;
                        if ( header.type != PED_PARTITION_EXTENDED && header.fsnameSize != 0 ) {
                                fsname = (char *) malloc(header.fsnameSize + 1);
                                fsname[header.fsnameSize] = '\0';
                                if ( readn(imaged, fsname, header.fsnameSize) != header.fsnameSize ) {
                                        perror("Reading filesystem name failed");
                                        return 1;
                                }

                                cerr << "Filesystem: " << fsname << endl;
                                free(fsname);
                        }

                        cerr << "Block size: " << header.blockSize << endl;
                        cerr << "Total blocks: " << header.blocksCount << endl;
                        cerr << "Used blocks: " << header.usedBlocks << endl;
                        cerr << "Bitmap allocated size: " << header.bitmapSize << endl;

                        //  skip the bitmap and the saved data
                        if ( lseek(imaged, header.bitmapSize, SEEK_CUR) == -1 ) {
                                perror("Skipping bitmap failed");
                                return 1;
                        }

                        if ( lseek(imaged, header.blockSize * header.usedBlocks, SEEK_CUR) == -1 ) {
                                perror("Skipping data failed");
                                return 1;
                        }
                }
        }
        else {
                // read the partition header
                PartitionHeader header;
                if ( readn(imaged, &header, sizeof(PartitionHeader)) != sizeof(PartitionHeader) ) {
                        perror("Reading partition header failed");
                        return 1;
                }

                // convert the header: image byte order -> cpu byte order
                // will be much easier to use
                header.type = img16_cpu(header.type);
                header.startSector = img64_cpu(header.startSector);
                header.endSector = img64_cpu(header.endSector);
                header.blockSize = img64_cpu(header.blockSize);
                header.blocksCount = img64_cpu(header.blocksCount);
                header.usedBlocks = img64_cpu(header.usedBlocks);
                header.fsnameSize = img16_cpu(header.fsnameSize);
                header.bitmapSize = img64_cpu(header.bitmapSize);
                header.bitmapOffset = img64_cpu(header.bitmapOffset);

                cerr << " == Partition: type " << header.type << ", geometry {"
                        << header.startSector << ", " << header.endSector << "} == " << endl;

                if ( header.type == PED_PARTITION_EXTENDED )
                        return 0;

                // read the filesystem name
                char *fsname = 0;
                if ( header.type != PED_PARTITION_EXTENDED && header.fsnameSize != 0 ) {
                        fsname = (char *) malloc(header.fsnameSize + 1);
                        fsname[header.fsnameSize] = '\0';
                        if ( readn(imaged, fsname, header.fsnameSize) != header.fsnameSize ) {
                                perror("Reading filesystem name failed");
                                return 1;
                        }

                        cerr << "Filesystem: " << fsname << endl;
                        free(fsname);
                }

                cerr << "Block size: " << header.blockSize << endl;
                cerr << "Total blocks: " << header.blocksCount << endl;
                cerr << "Used blocks: " << header.usedBlocks << endl;
                cerr << "Bitmap allocated size: " << header.bitmapSize << endl;
        }

        close(imaged);
        return 0;
}
