/*
    Copyright (C) 2007    Dmitry Ivanov, vonami@gmail.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <ntfs/volume.h>
#include <ntfs/attrib.h>

#include "ntfspartition_c.h"

void* partimage_ng_ntfs_open(const char *device)
{
        ntfs_volume *ntfs_vol = ntfs_mount(device, NTFS_MNT_RDONLY);
        if ( ntfs_vol == 0 ) {
                perror("Couldn't open the filesystem");
                return 0;
        }

        return ntfs_vol;
}

int partimage_ng_ntfs_close(void *ntfs_vol)
{
        if ( ntfs_vol == 0 ) {
                printf("Ntfs partition not opened yet\n");
                return errno;
        }

        if ( ntfs_umount((ntfs_volume *)ntfs_vol, TRUE) != 0 ) {
                perror("Couldn't close the filesystem");
                return errno;
        }

        return 0;
}

u_int8_t* partimage_ng_ntfs_bitmap(void *ntfs_vol)
{
        if ( ntfs_vol == 0 ) {
                printf("Ntfs partition not opened yet\n");
                return 0;
        }

        ntfs_attr *ntfs_bitmap = ((ntfs_volume *)ntfs_vol)->lcnbmp_na;

        u_int8_t *bitmap = (u_int8_t *) malloc(ntfs_bitmap->allocated_size);

        int64_t size = ntfs_attr_pread(ntfs_bitmap, 0, ntfs_bitmap->allocated_size, bitmap);
        if ( size != ntfs_bitmap->data_size ) {
                perror("ntfs_attr_pread failed");
                free(bitmap);
                return 0;
        }

        return bitmap;
}

u_int64_t partimage_ng_ntfs_bitmap_allocated_size(void *ntfs_vol)
{
        if ( ntfs_vol == 0 ) {
                printf("Ntfs partition not opened yet\n");
                return 0;
        }

        return ((ntfs_volume *)ntfs_vol)->lcnbmp_na->allocated_size;
}

u_int64_t partimage_ng_ntfs_bitmap_data_size(void *ntfs_vol)
{
        if ( ntfs_vol == 0 ) {
                printf("Ntfs partition not opened yet\n");
                return 0;
        }

        return ((ntfs_volume *)ntfs_vol)->lcnbmp_na->data_size;
}

size_t partimage_ng_ntfs_block_size(void *ntfs_vol)
{
        if ( ntfs_vol == 0 ) {
                printf("Ntfs partition not opened yet\n");
                return 0;
        }

        return ((ntfs_volume *)ntfs_vol)->cluster_size;
}

u_int64_t partimage_ng_ntfs_blocks_count(void *ntfs_vol)
{
        if ( ntfs_vol == 0 ) {
                printf("Ntfs partition not opened yet\n");
                return 0;
        }

        return ((ntfs_volume *)ntfs_vol)->nr_clusters;
}

struct ntfs_badblocks_run* partimage_ng_ntfs_badclusters(void *ntfs_vol)
{
        if ( ntfs_vol == 0 ) {
                printf("Ntfs partition not opened yet\n");
                return 0;
        }

        // find the $Bad attribute in the $BadClus MFT record
        ntfs_inode *ni;
        ntfs_attr_search_ctx *ctx = NULL;
        ntfschar *ustr;
        int len = 0;
        runlist *rl, *rl_bad;

        if ( (ni = ntfs_inode_open((ntfs_volume *)ntfs_vol, FILE_BadClus)) == NULL ) {
                printf("Failed to open the $BadClus record");
                return 0;
        }

        // allocate search contex
        if ( (ctx = ntfs_attr_get_search_ctx(ni, NULL)) == NULL ) {
                perror("Failed to allocate search context");
                ntfs_inode_close(ni);
                return 0;
        }

        // convert the name to unicode
        if ( (ustr = ntfs_str2ucs("$Bad", &len)) == NULL ) {
                perror("Couldn't convert '$Bad' to Unicode");
                ntfs_inode_close(ni);
                ntfs_attr_put_search_ctx(ctx);
                return 0;
        }

        // lookup the $Bad attribute
        if ( ntfs_attr_lookup(AT_DATA, ustr, len, CASE_SENSITIVE, 0, NULL, 0, ctx) != 0 ) {
                perror("Couldn't find the $Bad attribute");
                ntfs_inode_close(ni);
                ntfs_attr_put_search_ctx(ctx);
                ntfs_ucsfree(ustr);
                return 0;
        }

        // convert the mapping pairs to runlist
        if ( (rl_bad = ntfs_mapping_pairs_decompress((ntfs_volume *)ntfs_vol, ctx->attr, NULL)) == NULL ) {
                perror("Failed to decompress the mapping pairs to runlist");
                ntfs_inode_close(ni);
                ntfs_attr_put_search_ctx(ctx);
                ntfs_ucsfree(ustr);
                return 0;
        }

        // first pass: count up the number of non-empty runlists
        u_int32_t bbr_number = 0;
        for ( rl = rl_bad; rl->length; rl++ ) {
                if ( rl->lcn >= 0 )
                        bbr_number++;
        }

        // second pass: grab them
        struct ntfs_badblocks_run *bbr = (struct ntfs_badblocks_run *) malloc(
                                               (bbr_number + 1) * sizeof(struct ntfs_badblocks_run));
        bbr_number = 0;
        for ( rl = rl_bad; rl->length; rl++ ) {
                if ( rl->lcn >= 0 ) {
                        bbr[bbr_number].start = rl->lcn;
                        bbr[bbr_number].length = rl->length;
                        bbr_number++;
                }
        }

        // terminate the list
        bbr[bbr_number].start = 0;
        bbr[bbr_number].length = 0;

        // clean up
        ntfs_inode_close(ni);
        ntfs_attr_put_search_ctx(ctx);
        ntfs_ucsfree(ustr);
        free(rl_bad);

        return bbr;
}
