/*
    Copyright (C) 2007    Dmitry Ivanov, vonami@gmail.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cstring>
#include <iostream>

#include "partitionmanager.h"
#include "ext2partition.h"
#include "ntfspartition.h"
#include "fat32partition.h"
#include "linuxswappartition.h"
#include "unknownpartition.h"

using namespace PartimageNG;

bool PartitionManager::probed = false;
map<string, list<string> > PartitionManager::allDisks = map<string, list<string> >();
map<string, string> PartitionManager::allLabels = map<string, string>();
map<string, PedPartition *> PartitionManager::allPartitions = map<string, PedPartition *>();

void PartitionManager::probe()
{
        ped_device_probe_all();
        PedDevice *dev = 0;
        PedPartition *part = 0;
        do {
                dev = ped_device_get_next(dev);
                if ( !dev )
                        break;

                PedDisk *disk = 0;
                if ( (disk = ped_disk_new(dev)) == 0 ) {
                        // unrecognized disk label
                        // add to the list of drives and skip detecting partitions on it
                        string st(dev->path);
                        allLabels[st] = string("unknown");
                        allDisks[st] = list<string>();
                        continue;
                }
                string st(dev->path);
                allLabels[st] = string(disk->type->name);

                list<string> parts;
                do {
                        part = ped_disk_next_partition(disk, part);
                        if ( !part )
                                break;

                        if ( part->type == PED_PARTITION_NORMAL || part->type == PED_PARTITION_LOGICAL ||
                             part->type == PED_PARTITION_EXTENDED ) {
                                string st(ped_partition_get_path(part));
                                parts.push_back(st);
                                allPartitions[st] = part;
                        }
                } while (part);

                allDisks[st] = parts;

        } while ( dev );

        dump();
        probed = true;
}

void PartitionManager::dump()
{
        cerr << "Detected drives and partitions:" << endl;
        for ( map<string, list<string> >::iterator it_disk = allDisks.begin(); it_disk != allDisks.end(); it_disk++ ) {
                cerr << "Drive: " << it_disk->first << " { ";
                list<string> disklist = it_disk->second;
                for (list<string>::iterator it_part = disklist.begin(); it_part != disklist.end(); it_part++ ) {
                        cerr << *it_part << " ";
                }
                cerr << "}" << endl;
        }


        for ( map<string, string>::iterator it_label = allLabels.begin(); it_label != allLabels.end(); it_label++ ) {
                cerr << "Drive: " << it_label->first << ", label " << it_label->second << endl;
        }

        for ( map<string, PedPartition *>::iterator it_part = allPartitions.begin(); it_part != allPartitions.end(); it_part++ ) {
                cerr << "Partition: " << it_part->first << ", type " << it_part->second->type;
                if ( it_part->second->type != PED_PARTITION_EXTENDED && it_part->second->fs_type != 0 ) {
                        cerr << ", filesystem " << it_part->second->fs_type->name;
                }
                cerr << endl;
        }
}

int PartitionManager::createLabel(const char *diskpath, const char *labelname)
{
        if ( !probed )
                probe();

        // create a new partition table on the disk
        PedDevice *dev = 0;
        if ( (dev = ped_device_get(diskpath)) == 0 ) {
                cerr << "ped_device_get(): failed to open the device " << diskpath << endl;
                return 1;
        }

        PedDiskType *disktype = 0;
        if ( (disktype = ped_disk_type_get(labelname)) == 0 ) {
                cerr << "Cannot handle this partition table type: " << labelname << endl;
                return 1;
        }

        PedDisk *disk = 0;
        if ( (disk = ped_disk_new_fresh(dev, disktype)) == 0 ) {
                cerr << "Cannot create a new partition table on the device" << endl;
                return 1;
        }

        // commit changes
        if ( ped_disk_commit_to_dev(disk) == 0 ) {
                cerr << "Failed to commit changes to device" << endl;
                return 1;
        }
        if ( ped_disk_commit_to_os(disk) == 0 ) {
                cerr << "Failed to commit changes to operating system" << endl;
                return 1;
        }
        // sleep for a while (changes aren't applied immediately)
        sleep(2);

        // remove all its partitions
        list<string> parts = allDisks[string(diskpath)];
        for ( list<string>::iterator it_part = parts.begin(); it_part != parts.end(); it_part++ ) {
                allPartitions.erase(*it_part);
        }
        allDisks[string(diskpath)] = list<string>();

        // replace the label
        allLabels[string(diskpath)] = string(labelname);

        return 0;
}

const char* PartitionManager::createPartition(const char *diskpath, u_int16_t type,
                                              const char *fsname, u_int64_t startSector, u_int64_t endSector)
{
        if ( !probed )
                probe();

        PedDevice *dev = 0;
        if ( (dev = ped_device_get(diskpath)) == 0 ) {
                cerr << "ped_device_get(): failed to open the device " << diskpath << endl;
                return 0;
        }

        PedDisk *label = 0;
        if ( (label = ped_disk_new(dev)) == 0 ) {
                cerr << "Couldn't read the partition table" << endl;
                return 0;
        }

        PedFileSystemType *fstype = 0;
        if ( type != PED_PARTITION_EXTENDED ) {
                fstype = ped_file_system_type_get(fsname);
                if ( fstype == 0 ) {
                        cerr << "Cannot handle this filesystem type, sorry" << endl;
                }
        }

        // create and add a new partition
        PedPartition *newpart = ped_partition_new(label, (PedPartitionType) type, fstype, startSector, endSector);
        if ( newpart == 0 ) {
                cerr << "Creating partition failed" << endl;
                return 0;
        }

        if ( ped_disk_add_partition(label, newpart, ped_constraint_exact(&(newpart->geom))) == 0 ) {
                cerr << "Adding partition to the disk failed" << endl;
                return 0;
        }

        // commit changes
        if ( ped_disk_commit_to_dev(label) == 0 ) {
                cerr << "Failed to commit changes to device" << endl;
                return 0;
        }
        if ( ped_disk_commit_to_os(label) == 0 ) {
                cerr << "Failed to commit changes to operating system" << endl;
                return 0;
        }
        // sleep for a while (changes aren't applied immediately)
        sleep(2);

        // reload the partitions
        list<string> parts;
        PedPartition *part = 0;
        do {
                part = ped_disk_next_partition(label, part);
                if ( !part )
                        break;

                if ( part->type == PED_PARTITION_NORMAL || part->type == PED_PARTITION_LOGICAL ||
                        part->type == PED_PARTITION_EXTENDED ) {
                        string st(ped_partition_get_path(part));
                        parts.push_back(st);
                        allPartitions[st] = part;
                }
        } while (part);

        allDisks[string(diskpath)] = parts;

        return ped_partition_get_path(newpart);
}

bool PartitionManager::formatPartition(const char *partpath, const char *fsname)
{
        PedPartition *part = 0;
        map<string, PedPartition *>::iterator it_part = allPartitions.find(string(partpath));
        if ( it_part != allPartitions.end() ) {
                part = it_part->second;
        }
        else {
                cerr << "Cannot find this partition: " << partpath << endl;
                return false;
        }

        PedFileSystemType *fstype = 0;
        fstype = ped_file_system_type_get(fsname);
        if ( fstype == 0 ) {
                cerr << "Cannot handle this filesystem type, sorry" << endl;
                return false;
        }

        if ( ped_file_system_create(&part->geom, fstype, 0) == 0 ) {
                cerr << "Cannot format partition " << partpath << " as " << fsname << endl;
                return false;
        }

        return true;
}

const char* PartitionManager::labelname(const char *diskpath)
{
        if ( !probed )
                probe();

        map<string, string>::iterator it_label = allLabels.find(string(diskpath));
        if ( it_label != allLabels.end() )
                return it_label->second.c_str();
        else
                return 0;
}

const char* PartitionManager::filesystem(const char *partpath)
{
        if ( !probed )
                probe();

        PedPartition *part = 0;
        map<string, PedPartition *>::iterator it_part = allPartitions.find(string(partpath));
        if ( it_part != allPartitions.end() )
                part = it_part->second;
        else
                return 0;

        if ( part->type == PED_PARTITION_EXTENDED ) {
                return 0;
        }

        if ( part->fs_type == 0 || part->fs_type->name == 0 ) {
                return 0;
        }

        return part->fs_type->name;
}

list<string> PartitionManager::partitionsList(const char *diskpath)
{
        if ( !probed )
                probe();

        map<string, list<string> >::iterator it_disk = allDisks.find(string(diskpath));
        if ( it_disk != allDisks.end() )
                return it_disk->second;

        return list<string>();
}

Partition* PartitionManager::instantiatePartition(const char *partpath, InstantiateOption opt)
{
        if ( !probed )
                probe();

        PedPartition *part = 0;
        map<string, PedPartition *>::iterator it_part = allPartitions.find(string(partpath));
        if ( it_part != allPartitions.end() )
                part = it_part->second;
        else
                return 0;

        if ( opt == ForceUnknownFilesystem ) {
                return new UnknownPartition(part);
        }

        if ( part->type == PED_PARTITION_EXTENDED ) {
                return new UnknownPartition(part);
        }

        if ( part->fs_type == 0 || part->fs_type->name == 0 ) {
                return new UnknownPartition(part);
        }
        else if ( strcmp(part->fs_type->name, "ext2") == 0 ) {
                return new Ext2Partition(part);
        }
        else if ( strcmp(part->fs_type->name, "ext3") == 0 ) {
                return new Ext2Partition(part);
        }
        else if ( strcmp(part->fs_type->name, "ntfs") == 0 ) {
                return new NtfsPartition(part);
        }
        else if ( strcmp(part->fs_type->name, "fat32") == 0 ) {
                return new Fat32Partition(part);
        }
        else if ( strcmp(part->fs_type->name, "linux-swap") == 0 ) {
                return new LinuxswapPartition(part);
        }
        else {
                return new UnknownPartition(part);
        }

        return 0;
}

bool PartitionManager::isDrive(const char *drivepath)
{
        if ( !probed )
                probe();

        return ( allDisks.find(string(drivepath)) != allDisks.end() );
}

bool PartitionManager::isPartition(const char *partpath)
{
        if ( !probed )
                probe();

        return ( allPartitions.find(string(partpath)) != allPartitions.end() );
}

bool PartitionManager::isBusy(const char *devicepath)
{
        if ( !probed )
                probe();

        if ( isDrive(devicepath) ) {
                PedDevice *dev = ped_device_get(devicepath);
                return (ped_device_is_busy(dev) == 1 ? true : false);
        }
        else if ( isPartition(devicepath) ) {
                map<string, PedPartition *>::iterator it_part = allPartitions.find(string(devicepath));
                PedPartition *part = it_part->second;
                return (ped_partition_is_busy(part) == 1 ? true : false);
        }

        // silently return false
        // it's caller's responsibility to check beforehand
        // if this is a entire drive or a partition
        return false;
}
